"""
Shared list and metadata
"""

# stdlib
from functools import lru_cache
from typing import List

# module
from avwx.exceptions import BadStation
from avwx.load_utils import LazyLoad
from avwx.static.core import IN_REGIONS, M_IN_REGIONS, M_NA_REGIONS, NA_REGIONS

__LAST_UPDATED__ = "2021-04-23"

# Lazy data loading to speed up import times for unused features
STATIONS = LazyLoad("stations")

# maxsize = 2 ** number of boolean options
@lru_cache(maxsize=2)
def station_list(reporting: bool = True) -> List[str]:
    """Returns a list of station idents matching the search criteria"""
    stations = []
    for icao, station in STATIONS.items():
        if not reporting or station["reporting"]:
            stations.append(icao)
    return stations


def uses_na_format(station: str) -> bool:
    """Returns True if the station uses the North American format,

    False if the International format
    """
    if station[0] in NA_REGIONS:
        return True
    if station[0] in IN_REGIONS:
        return False
    if station[:2] in M_NA_REGIONS:
        return True
    if station[:2] in M_IN_REGIONS:
        return False
    raise BadStation("Station doesn't start with a recognized character set")


def valid_station(station: str):
    """Checks the validity of a station ident

    This function doesn't return anything. It merely raises a BadStation error if needed
    """
    station = station.strip()
    if len(station) != 4:
        raise BadStation("ICAO station ident must be four characters long")
    uses_na_format(station)



"""
Forecast report shared resources
"""

# stdlib
from datetime import datetime, timedelta, timezone
from typing import Callable, List, Optional, Tuple, Union

# module
from avwx.base import AVWXBase
from avwx.parsing import core
from avwx.service import Service
from avwx.structs import Code, Number, ReportData, Timestamp


def _split_line(
    line: str, size: int = 3, prefix: int = 4, strip: str = " |"
) -> List[str]:
    """Evenly split a string while stripping elements"""
    line = line[prefix:]
    ret = []
    while len(line) >= size:
        ret.append(line[:size].strip(strip))
        line = line[size:]
    line = line.strip(strip)
    if line:
        ret.append(line)
    return ret


def _timestamp(line: str) -> Timestamp:
    """Returns the report timestamp from the first line"""
    start = line.find("GUIDANCE") + 11
    text = line[start : start + 16].strip()
    timestamp = datetime.strptime(text, r"%m/%d/%Y  %H%M")
    return Timestamp(text, timestamp.replace(tzinfo=timezone.utc))


def _find_time_periods(line: List[str], timestamp: Optional[datetime]) -> List[dict]:
    """Find and create the empty time periods"""
    periods: List[Optional[Timestamp]] = []
    if timestamp is None:
        periods = [None] * len(line)
    else:
        previous = timestamp.hour
        for hourstr in line:
            if not hourstr:
                continue
            hour = int(hourstr)
            previous, difference = hour, hour - previous
            if difference < 0:
                difference += 24
            timestamp += timedelta(hours=difference)
            periods.append(Timestamp(hourstr, timestamp))
    return [{"time": time} for time in periods]


def _init_parse(report: str) -> Tuple[ReportData, List[str]]:
    """Returns the meta data and lines from a report string"""
    report = report.strip()
    lines = report.split("\n")
    struct = ReportData(
        raw=report,
        sanitized=report,
        station=report[:4],
        time=_timestamp(lines[0]),
        remarks=None,
    )
    return struct, lines


def _numbers(
    line: str,
    size: int = 3,
    prefix: str = "",
    postfix: str = "",
    decimal: int = None,
    literal: bool = False,
    special: dict = None,
) -> List[Optional[Number]]:
    """Parse line into Number objects

    Prefix, postfix, and decimal location are applied to value, not repr

    Decimal is applied after prefix and postfix
    """
    ret = []
    for item in _split_line(line, size=size):
        value = None
        if item:
            value = prefix + item + postfix
            if decimal is not None:
                if abs(decimal) > len(value):
                    value = value.zfill(abs(decimal))
                value = value[:decimal] + "." + value[decimal:]
        ret.append(core.make_number(value, repr=item, literal=literal, special=special))
    return ret


def _decimal_10(line: str, size: int = 3) -> List[Optional[Number]]:
    """Parse line into Number objects with 10ths decimal location"""
    return _numbers(line, size, decimal=-1)


def _decimal_100(line: str, size: int = 3) -> List[Optional[Number]]:
    """Parse line into Number objects with 100ths decimal location"""
    return _numbers(line, size, decimal=-2)


def _number_10(line: str, size: int = 3) -> List[Optional[Number]]:
    """Parse line into Number objects in tens"""
    return _numbers(line, size, postfix="0")


def _number_100(line: str, size: int = 3) -> List[Optional[Number]]:
    """Parse line into Number objects in hundreds"""
    return _numbers(line, size, postfix="00")


def _direction(line: str, size: int = 3) -> List[Optional[Number]]:
    """Parse line into Number objects in hundreds"""
    return _numbers(line, size, postfix="0", literal=True)


def _code(mapping: dict) -> Callable:
    """Generates a conditional code mapping function"""

    def func(line: str, size: int = 3) -> List[Union[Code, str, None]]:
        ret: List[Union[Code, str, None]] = []
        for key in _split_line(line, size=size):
            try:
                ret.append(Code(key, mapping[key]))
            except KeyError:
                ret.append(key or None)
        return ret

    return func


def _parse_lines(
    periods: List[dict],
    lines: List[str],
    handlers: Union[dict, Callable],
    size: int = 3,
):
    """Add data to time periods by parsing each line (element type)

    Adds data in place
    """
    for line in lines:
        try:
            key = line[:3]
            resp = handlers[key] if isinstance(handlers, dict) else handlers(key)
            *keys, handler = resp
        except (IndexError, KeyError):
            continue
        values = handler(line, size=size)
        values += [None] * (len(periods) - len(values))
        # pylint: disable=consider-using-enumerate
        for i in range(len(periods)):
            value = values[i]
            if not value:
                continue
            if isinstance(value, tuple):
                for j, key in enumerate(keys):
                    if value[j]:
                        periods[i][key] = value[j]
            else:
                periods[i][keys[0]] = value


class Forecast(AVWXBase):
    """Forecast base class"""

    # pylint: disable=abstract-method

    report_type: str
    _service_class: Service

    def __init__(self, icao: str):
        super().__init__(icao)
        self.service = self._service_class(self.report_type)  # type: ignore



"""
Classes for retrieving raw report strings via web scraping
"""

# pylint: disable=arguments-differ,invalid-name

# stdlib
import asyncio as aio
import random
from typing import Any, List, Optional, Tuple, Union

# library
from xmltodict import parse as parsexml  # type: ignore

# module
from avwx.parsing.core import dedupe
from avwx.exceptions import InvalidRequest
from avwx.station import valid_station
from avwx.service.base import CallsHTTP, Service


USER_AGENTS = [
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_6) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/14.0.3 Safari/605.1.15"
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_5) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/13.1.1 Safari/605.1.15",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:77.0) Gecko/20100101 Firefox/77.0",
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_5) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/83.0.4103.97 Safari/537.36",
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10.15; rv:77.0) Gecko/20100101 Firefox/77.0",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/83.0.4103.97 Safari/537.36",
]


class ScrapeService(Service, CallsHTTP):
    """Service class for fetching reports via direct web requests"""

    _valid_types: Tuple[str, ...] = ("metar", "taf")
    _strip_whitespace: bool = True

    def _make_err(self, body: str, key: str = "report path") -> InvalidRequest:
        """Returns an InvalidRequest exception with formatted error message"""
        msg = f"Could not find {key} in {self.__class__.__name__} response\n"
        return InvalidRequest(msg + body)

    @staticmethod
    def _make_headers() -> dict:
        """Returns request headers"""
        return {}

    # pylint: disable=unused-argument
    @staticmethod
    def _post_data(station: str) -> dict:
        """Returns the POST form/data payload"""
        return {}

    def _clean_report(self, report: Any) -> Any:
        """Replaces all *whitespace elements with a single space if enabled"""
        if not self._strip_whitespace:
            return report
        if isinstance(report, list):
            return dedupe(" ".join(r.split()) for r in report)
        if isinstance(report, str):
            return " ".join(report.split())
        return report


class StationScrape(ScrapeService):
    """Service class fetching reports from a station ICAO"""

    def _make_url(self, station: str) -> Tuple[str, dict]:
        """Returns a formatted URL and parameters"""
        raise NotImplementedError()

    def _extract(self, raw: str, station: str) -> str:
        """Extracts the report string from the service response"""
        raise NotImplementedError()

    async def _fetch(self, station: str, url: str, params: dict, timeout: int) -> str:
        headers = self._make_headers()
        data = self._post_data(station) if self.method.lower() == "post" else None
        text = await self._call(
            url, params=params, headers=headers, data=data, timeout=timeout
        )
        report = self._extract(text, station)
        return self._clean_report(report)

    def fetch(
        self,
        station: str,
        timeout: int = 10,
    ) -> str:
        """Fetches a report string from the service"""
        return aio.run(self.async_fetch(station, timeout))

    async def async_fetch(self, station: str, timeout: int = 10) -> str:
        """Asynchronously fetch a report string from the service"""
        valid_station(station)
        url, params = self._make_url(station)
        return await self._fetch(station, url, params, timeout)


# Multiple sources for NOAA data


class NOAA_ADDS(ScrapeService):
    """Requests data from NOAA ADDS"""

    url = "https://aviationweather.gov/adds/dataserver_current/httpparam"

    _valid_types = ("metar", "taf", "aircraftreport")
    _rtype_map = {"airep": "aircraftreport"}
    _targets = {"metar": "METAR", "taf": "TAF", "aircraftreport": "AircraftReport"}
    _coallate = ("aircraftreport",)

    def __init__(self, request_type: str):
        super().__init__(self._rtype_map.get(request_type, request_type))

    def _make_url(
        self, station: Optional[str], lat: Optional[float], lon: Optional[float]
    ) -> Tuple[str, dict]:
        """Returns a formatted URL and parameters"""
        # Base request params
        params = {
            "requestType": "retrieve",
            "format": "XML",
            "hoursBeforeNow": 2,
            "dataSource": self.report_type + "s",
        }
        if self.report_type == "aircraftreport":
            params["radialDistance"] = f"200;{lon},{lat}"
        else:
            params["stationString"] = station
        return self.url, params

    def _extract(self, raw: str) -> Union[str, List[str]]:
        """Extracts the raw_report element from XML response"""
        resp = parsexml(raw)
        try:
            data = resp["response"]["data"]
            if data["@num_results"] == "0":
                return ""
            reports = data[self._targets[self.report_type]]
        except KeyError as key_error:
            raise self._make_err(raw) from key_error
        # Only one report exists
        if isinstance(reports, dict):
            ret = reports["raw_text"]
            if self.report_type in self._coallate:
                ret = [ret]
        # Multiple reports exist
        elif isinstance(reports, list) and reports:
            if self.report_type in self._coallate:
                ret = [r["raw_text"] for r in reports]
            else:
                ret = reports[0]["raw_text"]
        # Something went wrong
        else:
            raise self._make_err(raw, '"raw_text"')
        return ret

    def fetch(
        self,
        station: str = None,
        lat: float = None,
        lon: float = None,
        timeout: int = 10,
    ) -> Union[str, List[str]]:
        """Fetches a report string from the service"""
        return aio.run(self.async_fetch(station, lat, lon, timeout))

    async def async_fetch(
        self,
        station: str = None,
        lat: float = None,
        lon: float = None,
        timeout: int = 10,
    ) -> Union[str, List[str]]:
        """Asynchronously fetch a report string from the service"""
        if station:
            valid_station(station)
        elif lat is None or lon is None:
            raise ValueError("No valid fetch parameters")
        url, params = self._make_url(station, lat, lon)
        text = await self._call(url, params=params, timeout=timeout)
        report = self._extract(text)
        return self._clean_report(report)


class NOAA_FTP(StationScrape):
    """Requests data from NOAA via FTP"""

    url = "https://tgftp.nws.noaa.gov/data/{}/{}/stations/{}.TXT"

    def _make_url(self, station: str) -> Tuple[str, dict]:
        """Returns a formatted URL and parameters"""
        root = "forecasts" if self.report_type == "taf" else "observations"
        return self.url.format(root, self.report_type, station), {}

    def _extract(self, raw: str, station: str) -> str:
        """Extracts the report using string finding"""
        raw = raw[raw.find(station) :]
        return raw[: raw.find('"')]


class NOAA_Scrape(StationScrape):
    """Requests data from NOAA via site scraping"""

    url = "https://aviationweather.gov/{}/data"

    def _make_url(self, station: str) -> Tuple[str, dict]:
        """Returns a formatted URL and parameters"""
        hours = 7 if self.report_type == "taf" else 2
        return (
            self.url.format(self.report_type),
            {"ids": station, "format": "raw", "hours": hours},
        )

    def _extract(self, raw: str, station: str) -> str:
        """Extracts the report using string finding"""
        raw = raw[raw.find("<code>") :]
        raw = raw[raw.find(station) :]
        raw = raw[: raw.find("</code>")]
        for char in ("<br/>", "&nbsp;"):
            raw = raw.replace(char, " ")
        return raw


NOAA = NOAA_Scrape


# Regional data sources


class AMO(StationScrape):
    """Requests data from AMO KMA for Korean stations"""

    url = "http://amoapi.kma.go.kr/amoApi/{}"

    def _make_url(self, station: str) -> Tuple[str, dict]:
        """Returns a formatted URL and parameters"""
        return self.url.format(self.report_type), {"icao": station}

    def _extract(self, raw: str, station: str) -> str:
        """Extracts the report message from XML response"""
        resp = parsexml(raw)
        try:
            report = resp["response"]["body"]["items"]["item"][
                self.report_type.lower() + "Msg"
            ]
        except KeyError as key_error:
            raise self._make_err(raw) from key_error
        if not report:
            raise self._make_err("The station might not exist")
        # Replace line breaks
        report = report.replace("\n", "")
        # Remove excess leading and trailing data
        for item in (self.report_type.upper(), "SPECI"):
            if report.startswith(item + " "):
                report = report[len(item) + 1 :]
        report = report.rstrip("=")
        # Make every element single-spaced and stripped
        return " ".join(report.split())


class MAC(StationScrape):
    """Requests data from Meteorologia Aeronautica Civil for Columbian stations"""

    url = "http://meteorologia.aerocivil.gov.co/expert_text_query/parse"
    method = "POST"

    def _make_url(self, station: str) -> Tuple[str, dict]:
        """Returns a formatted URL and parameters"""
        return self.url, {"query": f"{self.report_type} {station}"}

    def _extract(self, raw: str, station: str) -> str:
        """Extracts the report message using string finding"""
        report = raw[raw.find(station.upper() + " ") :]
        report = report[: report.find(" =")]
        return report


class AUBOM(StationScrape):
    """Requests data from the Australian Bureau of Meteorology"""

    url = "http://www.bom.gov.au/aviation/php/process.php"
    method = "POST"

    def _make_url(self, _) -> Tuple[str, dict]:
        """Returns a formatted URL and empty parameters"""
        return self.url, {}

    @staticmethod
    def _make_headers() -> dict:
        """Returns request headers"""
        return {
            "Content-Type": "application/x-www-form-urlencoded",
            "Accept": "*/*",
            "Accept-Language": "en-us",
            "Accept-Encoding": "gzip, deflate",
            "Host": "www.bom.gov.au",
            "Origin": "http://www.bom.gov.au",
            "User-Agent": random.choice(USER_AGENTS),
            "Connection": "keep-alive",
            # 'Referer': 'http://www.bom.gov.au/aviation/observations/metar-speci/',
            # 'Cookie': 'check=ok; bm_sv=C423C389DECFF9A821E265C1696BFD2B~HNlY4KT2eVGRmQ7vlgQu8I6Uj0MQIueL8syy1Gce2Do+dyNtfywMIueEu+bDc5gEsw2QMkRJY16RvBDBYgrCZJRlY08iB6teeIZuYaWHzqIJOWEXJANCWC6HgGu4uUEOvVRmMWFiraXWm3YTUv38sNxB83MXOZiS5YS9qf6MqXo=; __utma=172860464.760688564.1618331786.1618331786.1618331786.1; __utmb=172860464.4.10.1618331786; __utmc=172860464; __utmz=172860464.1618331786.1.1.utmcsr=(direct)|utmccn=(direct)|utmcmd=(none); PHPSESSID=sbu75i840376vc6v33k8f7tp13; ak_bmsc=2DE29B1D2EEC759732EFC20AF6CE52AA1749B466715E000089C8756048C6D311~plkVCUVDvoBcyCISZlrLApGC8CbD9JMTgKXEBQqxfXzP039W0VONAXG4azzOIcDVp1+ovscHFBt1zxXTqr8sEJhqGS3i9Q2aBCg+hmKXYuJ5pmj0jiz8YTy+uoNui4FKcDiw2sGtzXMkv022BuUddJIj2SNSXNtt6ogjVtUeu1uD7qEp+hBkusmT9C2byp34EkVilnlIomV/n2svH3u7P8rk2sRXlcqjKp/hPP1nDly4EEjCv5VVFhw1JbVbxbw2v7; bm_mi=F129320EBC30F70FC3BCC129E5FB67B1~2ErkxZx/0YwGSQpO9GUziKtgD05/7Zl225D53UFU2Su+OzO9YwYEG/CGHTiTTQL9bNMHZjqYzRuszGROuGts+JHoICZx8rhM9O5BksHCw+xaBIIzhu0hCDMYYvGUgbSwc1PwdIzdgi6U4MSkWBxyOwN1Ol9WR9yfb84rxfdXXAfFpVIzir8rmw9XryJSgPG3FqFLXZCFt5ygWB+JLoUo/qksC/FHIgAnXhJHM5JvTj+oTdB91O2PbYUNHCbGR2pfSHO1pCOTaN5K8EQqEHKsOGgXMtqUjpes0Xx4B1817T4=',
            # 'X-Requested-With': 'XMLHttpRequest',
        }

    @staticmethod
    def _post_data(station: str) -> dict:
        """Returns the POST form"""
        return {"keyword": station, "type": "search", "page": "TAF"}

    def _extract(self, raw: str, station: str) -> str:
        """Extracts the reports from HTML response"""
        index = 1 if self.report_type == "taf" else 2
        try:
            report = raw.split("<p")[index]
            report = report[report.find(">") + 1 :]
        except IndexError as index_error:
            raise self._make_err("The station might not exist") from index_error
        if report.startswith("<"):
            return ""
        report = report[: report.find("</p>")]
        return report.replace("<br />", " ")


PREFERRED = {"RK": AMO, "SK": MAC}
BY_COUNTRY = {"AU": AUBOM}


def get_service(station: str, country_code: str) -> ScrapeService:
    """Returns the preferred service for a given station"""
    for prefix in PREFERRED:
        if station.startswith(prefix):
            return PREFERRED[prefix]  # type: ignore
    return BY_COUNTRY.get(country_code, NOAA)  # type: ignore



"""
Contains avwx custom exceptions
"""


class BadStation(Exception):
    """Station does not exist"""


class InvalidRequest(Exception):
    """Unable to fetch data"""


class SourceError(Exception):
    """Source servers returned an error code"""



"""
Contains report sanitization functions
"""

# stdlib
import re
from itertools import permutations
from typing import List, Optional

# module
from avwx.parsing.core import dedupe, is_timerange, is_timestamp, is_unknown
from avwx.static.core import CLOUD_LIST, CLOUD_TRANSLATIONS
from avwx.static.taf import TAF_NEWLINE, TAF_NEWLINE_STARTSWITH

STR_REPL = {
    " C A V O K ": " CAVOK ",
    "?": " ",
    '"': "",
    "'": "",
    "`": "",
    ".": "",
    " VTB": " VRB",
    " VBR": " VRB",
    " VBB": " VRB",
    " ERB": " VRB",
    " VRV": " VRB",
    " BRV": " VRB",
    " VRN": " VRB",
    " BRB": " VRB",
    " VEB": " VRB",
    " VFB": " VRB",
    " VAB": " VRB",
    " VAR": " VRB",
    " VRG": " VRB",
    " VRR": " VRB",
    " VGB": " VRB",
    " VRC": " VRB",
    " VRE": " VRB",
    " WBB": " VRB",
    " NRB": " VRB",
    " VRBN": " VRB",
    " VR0": " VRB0",
    " VB0": " VRB0",
    " RB0": " VRB0",
    " V0": " VRB0",
    " 0I0": " 090",
    " PROBB": " PROB",
    " PROBN": " PROB",
    # "Z/ ": "Z ", NOTE: Too broad re pirep
    "NOSIGKT ": "KT NOSIG ",
    "KTKT ": "KT ",
    "KKT ": "KT ",
    "KLT ": "KT ",
    "TKT ": "KT ",
    "GKT ": "KT ",
    " TMM": " TNM",
    " TNTN": " TN",
    " TXTX": " TX",
    "CALMKT ": "CALM ",
    "N0SIG": "NOSIG",
    " <1/": " M1/",  # <1/4SM <1/8SM
    "/04SM": "/4SM",
    "/08SM": "/8SM",
    " /34SM": "3/4SM",
}


def sanitize_report_string(txt: str) -> str:
    """Provides sanitization for operations that work better when the report is a string

    Returns the first pass sanitized report string
    """
    txt = txt.upper()
    if len(txt) < 4:
        return txt
    # Standardize whitespace
    txt = " ".join(txt.split())
    # Prevent changes to station ID
    stid, txt = txt[:4], txt[4:]
    # Replace invalid key-value pairs
    for key, rep in STR_REPL.items():
        txt = txt.replace(key, rep)
    # Check for missing spaces in front of cloud layers
    # Ex: TSFEW004SCT012FEW///CBBKN080
    for cloud in CLOUD_LIST:
        if cloud in txt and " " + cloud not in txt:
            start, counter = 0, 0
            while txt.count(cloud) != txt.count(" " + cloud):
                cloud_index = start + txt[start:].find(cloud)
                if len(txt[cloud_index:]) >= 3:
                    target = txt[
                        cloud_index + len(cloud) : cloud_index + len(cloud) + 3
                    ]
                    if target.isdigit() or not target.strip("/"):
                        txt = txt[:cloud_index] + " " + txt[cloud_index:]
                start = cloud_index + len(cloud) + 1
                # Prevent infinite loops
                if counter > txt.count(cloud):
                    break
                counter += 1
    return stid + txt


def extra_space_exists(str1: str, str2: str) -> bool:
    """Return True if a space shouldn't exist between two items"""
    ls1, ls2 = len(str1), len(str2)
    if str1.isdigit():
        # 10 SM
        if str2 in ["SM", "0SM"]:
            return True
        # 12 /10
        if ls2 > 2 and str2[0] == "/" and str2[1:].isdigit():
            return True
    if str2.isdigit():
        # OVC 040
        if str1 in CLOUD_LIST:
            return True
        # 12/ 10
        if ls1 > 2 and str1.endswith("/") and str1[:-1].isdigit():
            return True
        # 12/1 0
        if (
            ls2 == 1
            and ls1 > 3
            and str1[:2].isdigit()
            and "/" in str1
            and str1[3:].isdigit()
        ):
            return True
        # Q 1001
        if str1 in ["Q", "A"]:
            return True
    # 36010G20 KT
    if (
        str2 == "KT"
        and str1[-1].isdigit()
        and (str1[:5].isdigit() or (str1.startswith("VRB") and str1[3:5].isdigit()))
    ):
        return True
    # 36010K T
    if (
        str2 == "T"
        and ls1 >= 6
        and (str1[:5].isdigit() or (str1.startswith("VRB") and str1[3:5].isdigit()))
        and str1[-1] == "K"
    ):
        return True
    # OVC022 CB
    if (
        str2 in CLOUD_TRANSLATIONS
        and str2 not in CLOUD_LIST
        and ls1 >= 3
        and str1[:3] in CLOUD_LIST
    ):
        return True
    # FM 122400
    if str1 in ["FM", "TL"] and (
        str2.isdigit() or (str2.endswith("Z") and str2[:-1].isdigit())
    ):
        return True
    # TX 20/10
    if str1 in ["TX", "TN"] and str2.find("/") != -1:
        return True
    return False


_CLOUD_GROUP = "(" + "|".join(CLOUD_LIST) + ")"
CLOUD_SPACE_PATTERNS = [
    re.compile(pattern)
    for pattern in (
        r"(?=.+)" + _CLOUD_GROUP + r"\d{3}(\w{2,3})?$",  # SCT010BKN021
        r"M?\d{2}\/M?\d{2}$",  # BKN01826/25
    )
]


def extra_space_needed(item: str) -> Optional[int]:
    """Returns the index where the string should be separated or None"""
    # For items starting with cloud list
    if item[:3] in CLOUD_LIST:
        for pattern in CLOUD_SPACE_PATTERNS:
            match = pattern.search(item)
            if match is None:
                continue
            if match.start():
                return match.start()
    # Connected timestamp
    for loc, check in ((7, is_timestamp), (9, is_timerange)):
        if len(item) > loc and check(item[:loc]):
            return loc
    # Connected to wind
    if len(item) > 5 and "KT" in item and not item.endswith("KT"):
        index = item.find("KT")
        if index > 4:
            return index + 2
    # TAF newline connected to previous element
    for key in TAF_NEWLINE:
        if key in item and not item.startswith(key):
            return item.find(key)
    for key in TAF_NEWLINE_STARTSWITH:
        if key in item and not item.startswith(key):
            index = item.find(key)
            if item[index + len(key) :].isdigit():
                return index
    return None


ITEM_REMV = [
    "AUTO",
    "COR",
    "NSC",
    "NCD",
    "$",
    "KT",
    "M",
    ".",
    "RTD",
    "SPECI",
    "METAR",
    "CORR",
    "TTF",
]
ITEM_REPL = {"CALM": "00000KT"}
VIS_PERMUTATIONS = ["".join(p) for p in permutations("P6SM")]
VIS_PERMUTATIONS.remove("6MPS")


def sanitize_report_list(
    wxdata: List[str], remove_clr_and_skc: bool = True
) -> List[str]:
    """Sanitize wxData

    We can remove and identify "one-off" elements and fix other issues before parsing a line
    """
    for i, item in reversed(list(enumerate(wxdata))):
        ilen = len(item)
        # Remove elements containing only '/'
        if is_unknown(item):
            wxdata.pop(i)
            continue
        # Remove empty wind /////KT
        if item.endswith("KT") and is_unknown(item[:-2]):
            wxdata.pop(i)
            continue
        # Remove RE from wx codes, REVCTS -> VCTS
        if ilen in [4, 6] and item.startswith("RE"):
            wxdata[i] = item[2:]
        # Fix a slew of easily identifiable conditions where a space does not belong
        elif i and extra_space_exists(wxdata[i - 1], item):
            wxdata[i - 1] += wxdata.pop(i)
        # Remove spurious elements
        elif item in ITEM_REMV:
            wxdata.pop(i)
        # Remove 'Sky Clear' from METAR but not TAF
        elif remove_clr_and_skc and item in ["CLR", "SKC"]:
            wxdata.pop(i)
        # Replace certain items
        elif item in ITEM_REPL:
            wxdata[i] = ITEM_REPL[item]
        # Remove amend signifier from start of report ('CCA', 'CCB',etc)
        elif ilen == 3 and item.startswith("CC") and item[2].isalpha():
            wxdata.pop(i)
        # Fix inconsistent 'P6SM' Ex: TP6SM or 6PSM -> P6SM
        elif ilen > 3 and item[-4:] in VIS_PERMUTATIONS:
            wxdata[i] = "P6SM"
        # Fix misplaced KT 22022KTG40
        elif ilen == 10 and "KTG" in item and item[:5].isdigit():
            wxdata[i] = item.replace("KTG", "G") + "KT"
        # Fix malformed KT Ex: 06012G22TK
        if (
            ilen >= 7
            and (item[:3].isdigit() or item[:3] == "VRB")
            and item[-2:] in ("TK", "GT")
        ):
            wxdata[i] = item[:-2] + "KT"
        # Fix gust double G Ex: 360G17G32KT
        elif ilen > 10 and item.endswith("KT") and item[3] == "G":
            wxdata[i] = item[:3] + item[4:]
        # Fix leading character mistypes in wind
        elif (
            ilen > 7
            and not item[0].isdigit()
            and not item.startswith("VRB")
            and item.endswith("KT")
            and not item.startswith("WS")
        ):
            while not item[0].isdigit() and not item.startswith("VRB"):
                item = item[1:]
            wxdata[i] = item
        # Fix non-G gust Ex: 14010-15KT
        elif ilen == 10 and item.endswith("KT") and item[5] != "G":
            wxdata[i] = item[:5] + "G" + item[6:]
        # Fix leading digits on VRB wind Ex: 2VRB02KT
        elif (
            ilen > 7
            and item.endswith("KT")
            and "VRB" in item
            and item[0].isdigit()
            and "Z" not in item
        ):
            while item[0].isdigit():
                item = item[1:]
            wxdata[i] = item
        # Fix wind T
        elif not item.endswith("KT") and (
            (
                ilen == 6
                and item[5] in ["K", "T"]
                and (
                    item[:5].isdigit()
                    or (item.startswith("VRB") and item[:3].isdigit())
                )
            )
            or (
                ilen == 9
                and item[8] in ["K", "T"]
                and item[5] == "G"
                and (item[:5].isdigit() or item.startswith("VRB"))
            )
        ):
            wxdata[i] = item[:-1] + "KT"
        # Fix joined TX-TN
        elif ilen > 16 and len(item.split("/")) == 3:
            if item.startswith("TX") and "TN" not in item:
                tn_index = item.find("TN")
                wxdata.insert(i + 1, item[:tn_index])
                wxdata[i] = item[tn_index:]
            elif item.startswith("TN") and item.find("TX") != -1:
                tx_index = item.find("TX")
                wxdata.insert(i + 1, item[:tx_index])
                wxdata[i] = item[tx_index:]
        # Fix situations where a space is missing
        index = extra_space_needed(item)
        if index:
            wxdata.insert(i + 1, item[index:])
            wxdata[i] = item[:index]
    wxdata = dedupe(wxdata, only_neighbors=True)
    return wxdata



"""
Station text-based search
"""

# stdlib
from functools import lru_cache
from typing import Iterable, List, Tuple

# library
from rapidfuzz import fuzz, process  # type: ignore

# module
from avwx.load_utils import LazyCalc
from avwx.station.meta import STATIONS
from avwx.station.station import Station, station_filter


TYPE_ORDER = [
    "large_airport",
    "medium_airport",
    "small_airport",
    "seaplane_base",
    "heliport",
    "balloonport",
    "weather_station",
]


def _format_search(airport: dict, keys: Iterable[str]) -> str:
    values = [airport.get(k) for k in keys]
    return " - ".join(k for k in values if k)


def _build_corpus() -> List[str]:
    keys = ("icao", "iata", "city", "state", "name")
    return [_format_search(s, keys) for s in STATIONS.values()]


_CORPUS = LazyCalc(_build_corpus)


def _sort_key(result: Tuple[Station, int]) -> Tuple[int, ...]:
    station, score = result
    try:
        type_order = TYPE_ORDER.index(station.type)
    except ValueError:
        type_order = 10
    return (score, 10 - type_order)


@lru_cache(maxsize=128)
def search(
    text: str,
    limit: int = 10,
    is_airport: bool = False,
    sends_reports: bool = True,
) -> List[Station]:
    """Text search for stations against codes, name, city, and state

    Results may be shorter than limit value
    """
    results = process.extract(
        text, _CORPUS.value, limit=limit * 20, scorer=fuzz.token_set_ratio
    )
    results = [(Station.from_icao(k[:4]), s) for k, s, _ in results]
    results.sort(key=_sort_key, reverse=True)
    results = [s for s, _ in results if station_filter(s, is_airport, sends_reports)]
    return results[:limit] if len(results) > limit else results



"""
METAR data translation handlers
"""

import avwx.parsing.translate.base as _trans
from avwx.parsing import remarks
from avwx.structs import MetarData, MetarTrans, Units


def translate_metar(wxdata: MetarData, units: Units) -> MetarTrans:
    """Returns translations for a MetarData object"""
    shared = _trans.current_shared(wxdata, units)
    return MetarTrans(
        altimeter=shared.altimeter,
        clouds=shared.clouds,
        visibility=shared.visibility,
        wx_codes=shared.wx_codes,
        wind=_trans.wind(
            wxdata.wind_direction,
            wxdata.wind_speed,
            wxdata.wind_gust,
            wxdata.wind_variable_direction,
            units.wind_speed,
        ),
        temperature=_trans.temperature(wxdata.temperature, units.temperature),
        dewpoint=_trans.temperature(wxdata.dewpoint, units.temperature),
        remarks=remarks.translate(wxdata.remarks),
    )
